"""
Hyperbolic Householder transformations for Up- and Downdating Cholesky factorizations.
"""

__version__ = "1.0.2a3"

from ._hyhound import *  # noqa: F403
from ._hyhound import __version__ as __c_version__  # noqa: F401



import hyhound
import numpy as np
import numpy.linalg as la


def test_update_cholesky():
    rng = np.random.default_rng(seed=123)
    m, n = 7, 13
    L = np.tril(rng.uniform(-2, 2, (n, n)))
    A = rng.uniform(-1, 1, (n, m))
    L̃, Bu, Wu = hyhound.update_cholesky(L, A)
    assert la.norm(L @ L.T + A @ A.T - L̃ @ L̃.T, "fro") < 1e-12
    L2, Bd, Wd = hyhound.downdate_cholesky(L̃, A)
    assert la.norm(L @ L.T - L2 @ L2.T, "fro") < 1e-12

    L3, Au = hyhound.update_apply_householder(L, A, Bu, Wu)
    assert la.norm(Au, "fro") < 1e-11
    assert la.norm(L̃ - L3, "fro") < 1e-12
    L4, Ad = hyhound.downdate_apply_householder(L̃, A, Bd, Wd)
    assert la.norm(Ad, "fro") < 1e-11
    assert la.norm(L2 - L4, "fro") < 1e-12


def test_update_cholesky_tall():
    rng = np.random.default_rng(seed=123)
    m, n, p = 7, 13, 43
    L = np.tril(rng.uniform(-2, 2, (p, n)))
    A = rng.uniform(-1, 1, (p, m))
    L̃, Ã, Wu = hyhound.update_cholesky(L, A)
    Bu = Ã[:n, :].copy()
    Ã[:n, :] = 0
    assert la.norm(L @ L.T + A @ A.T - L̃ @ L̃.T - Ã @ Ã.T, "fro") < 1e-12
    L2, A2, Wd = hyhound.downdate_cholesky(L̃, A)
    Bd = A2[:n, :].copy()
    A2[:n, :] = 0
    assert la.norm(L @ L.T - L2 @ L2.T - Ã @ Ã.T + A2 @ A2.T, "fro") < 1e-12

    L3, Au = hyhound.update_apply_householder(L, A, Bu, Wu)
    assert la.norm(Au[:n, :], "fro") < 1e-10
    assert la.norm(L̃ - L3, "fro") < 1e-11
    L4, Ad = hyhound.downdate_apply_householder(L̃, A, Bd, Wd)
    assert la.norm(Ad[:n, :], "fro") < 1e-10
    assert la.norm(L2 - L4, "fro") < 1e-11


def test_update_cholesky_inplace():
    rng = np.random.default_rng(seed=123)
    m, n = 7, 13
    L = np.tril(rng.uniform(-2, 2, (n, n)))
    A = rng.uniform(-1, 1, (n, m))
    hyhound.update_cholesky_inplace(L̃ := L.copy(order="F"), A.copy(order="F"))
    assert la.norm(L @ L.T + A @ A.T - L̃ @ L̃.T, "fro") < 1e-12
    hyhound.downdate_cholesky_inplace(L̃, A.copy(order="F"))
    assert la.norm(L @ L.T - L̃ @ L̃.T, "fro") < 1e-12


def test_update_cholesky_diag():
    rng = np.random.default_rng(seed=123)
    m, n = 7, 13
    L = np.tril(rng.uniform(-2, 2, (n, n)))
    L += 10 * np.eye(n)
    A = rng.uniform(-1, 1, (n, m))
    d = rng.uniform(-1, 1, (m,))
    L̃, Bu, Wu = hyhound.update_cholesky_diag(L, A, d)
    assert la.norm(L @ L.T + A @ np.diag(d) @ A.T - L̃ @ L̃.T, "fro") < 1e-12
    L2, Bd, Wd = hyhound.update_cholesky_diag(L̃, A, -d)
    assert la.norm(L @ L.T - L2 @ L2.T, "fro") < 1e-12

    L3, Au = hyhound.update_apply_householder_diag(L, A, d, Bu, Wu)
    assert la.norm(Au, "fro") < 1e-11
    assert la.norm(L̃ - L3, "fro") < 1e-12
    L4, Ad = hyhound.update_apply_householder_diag(L̃, A, -d, Bd, Wd)
    assert la.norm(Ad, "fro") < 1e-11
    assert la.norm(L - L4, "fro") < 1e-12


def test_update_cholesky_sign():
    rng = np.random.default_rng(seed=123)
    m, n = 7, 13
    L = np.tril(rng.uniform(-2, 2, (n, n)))
    L += 10 * np.eye(n)
    A = rng.uniform(-1, 1, (n, m))
    d = rng.uniform(-1, 1, (m,))
    s = 0 * d  # np.copysign(np.zeros((m,)), d)
    e = np.copysign(np.ones((m,)), d)
    L̃, Bu, Wu = hyhound.update_cholesky_sign(L, A, s)
    assert la.norm(L @ L.T + A @ np.diag(e) @ A.T - L̃ @ L̃.T, "fro") < 1e-12
    L2, Bd, Wd = hyhound.update_cholesky_sign(L̃, A, -s)
    assert la.norm(L @ L.T - L2 @ L2.T, "fro") < 1e-12

    L3, Au = hyhound.update_apply_householder_sign(L, A, s, Bu, Wu)
    assert la.norm(Au, "fro") < 1e-11
    assert la.norm(L̃ - L3, "fro") < 1e-12
    L4, Ad = hyhound.update_apply_householder_sign(L̃, A, -s, Bd, Wd)
    assert la.norm(Ad, "fro") < 1e-11
    assert la.norm(L - L4, "fro") < 1e-12



from typing import Annotated, overload

import numpy
from numpy.typing import NDArray


build_time: str = '2025-11-02T15:27:37Z'

commit_hash: str = 'ed20e5be2b56ab49a71278082c731930100799b8'

@overload
def update_cholesky_inplace(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')]) -> None:
    """
    Cholesky factorization update. Overwrites its arguments.

    L̃L̃ᵀ + ÃÃᵀ = LLᵀ + AAᵀ

    Parameters
    ----------
    L : (k × n), lower-trapezoidal, Fortran order
        On entry, the original Cholesky factor L.
        On exit, contains the updated Cholesky factor L̃.

    A : (k × m), rectangular, Fortran order
        On entry, the update matrix A.
        On exit, contains the k-n bottom rows of the remaining update matrix Ã
        (the top n rows of Ã are zero).
    """

@overload
def update_cholesky_inplace(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')]) -> None: ...

@overload
def downdate_cholesky_inplace(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')]) -> None:
    """
    Cholesky factorization downdate. Overwrites its arguments.

    L̃L̃ᵀ - ÃÃᵀ = LLᵀ - AAᵀ

    Parameters
    ----------
    L : (k × n), lower-trapezoidal, Fortran order
        On entry, the original Cholesky factor L.
        On exit, contains the updated Cholesky factor L̃.

    A : (k × m), rectangular, Fortran order
        On entry, the downdate matrix A.
        On exit, contains the k-n bottom rows of the remaining downdate matrix Ã
        (the top n rows of Ã are zero).
    """

@overload
def downdate_cholesky_inplace(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')]) -> None: ...

@overload
def update_cholesky_sign_inplace(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], signs: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> None:
    """
    Cholesky factorization update with signed columns. Overwrites its arguments.

    L̃L̃ᵀ + ÃSÃᵀ = LLᵀ + ASAᵀ,
    where S = np.diag(np.copysign(np.ones(m), signs)) and signs contains ±0.

    Parameters
    ----------
    L : (k × n), lower-trapezoidal, Fortran order
        On entry, the original Cholesky factor L.
        On exit, contains the updated Cholesky factor L̃.

    A : (k × m), rectangular, Fortran order
        On entry, the update matrix A.
        On exit, contains the k-n bottom rows of the remaining update matrix Ã
        (the top n rows of Ã are zero).

    signs : m-vector
        Signs that determine whether a column of A is added (+0) or removed (-0).
        Values other than ±0 are not allowed.
    """

@overload
def update_cholesky_sign_inplace(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], signs: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> None: ...

@overload
def update_cholesky_diag_inplace(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu')], diag: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> None:
    """
    Cholesky factorization update with diagonal scaling. Overwrites its arguments.

    L̃L̃ᵀ + ÃDÃᵀ = LLᵀ + ADAᵀ,
    where D = np.diag(diag).

    Parameters
    ----------
    L : (k × n), lower-trapezoidal, Fortran order
        On entry, the original Cholesky factor L.
        On exit, contains the updated Cholesky factor L̃.

    A : (k × m), rectangular, Fortran order
        On entry, the update matrix A.
        On exit, contains the k-n bottom rows of the remaining update matrix Ã
        (the top n rows of Ã are zero).

    diag : m-vector
        Scale factors corresponding to the columns of A.
    """

@overload
def update_cholesky_diag_inplace(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu')], diag: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> None: ...

@overload
def update_cholesky(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Cholesky factorization update. Returns updated copies.

    L̃L̃ᵀ + ÃÃᵀ = LLᵀ + AAᵀ

    Parameters
    ----------
    L : (k × n), lower-trapezoidal
        The original Cholesky factor.

    A : (k × m), rectangular
        The update matrix.

    Returns
    -------
    L̃ : (k × n)
        The updated Cholesky factor.

    A_rem : (k × m)
        Contains the k-n bottom rows of the remaining update matrix Ã.
        The top n rows of Ã are zero (not stored explicitly).
        The top n rows of A_rem contain Householder reflectors.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update. Together with the top n rows of A_rem,
        this can be used to apply the block Householder transformation to other matrices.
        The number of rows depends on the block size and is architecture-dependent.
    """

@overload
def update_cholesky(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def downdate_cholesky(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Cholesky factorization downdate. Returns updated copies.

    L̃L̃ᵀ - ÃÃᵀ = LLᵀ - AAᵀ

    Parameters
    ----------
    L : (k × n), lower-trapezoidal
        The original Cholesky factor.

    A : (k × m), rectangular
        The downdate matrix.

    Returns
    -------
    L̃ : (k × n)
        The updated Cholesky factor.

    A_rem : (k × m)
        Contains the k-n bottom rows of the remaining downdate matrix Ã.
        The top n rows of Ã are zero (not stored explicitly).
        The top n rows of A_rem contain Householder reflectors.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update. Together with the top n rows of A_rem,
        this can be used to apply the block Householder transformation to other matrices.
        The number of rows depends on the block size and is architecture-dependent.
    """

@overload
def downdate_cholesky(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def update_cholesky_sign(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], signs: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Cholesky factorization update with signed columns. Returns updated copies.

    L̃L̃ᵀ + ÃSÃᵀ = LLᵀ + ASAᵀ,
    where S = np.diag(np.copysign(np.ones(m), signs)) and signs contains ±0.

    Parameters
    ----------
    L : (k × n), lower-trapezoidal
        The original Cholesky factor.

    A : (k × m), rectangular
        The update matrix.

    signs : m-vector
        Signs that determine whether a column of A is added (+0) or removed (-0).
        Values other than ±0 are not allowed.

    Returns
    -------
    L̃ : (k × n)
        The updated Cholesky factor.

    A_rem : (k × m)
        Contains the k-n bottom rows of the remaining update matrix Ã.
        The top n rows of Ã are zero (not stored explicitly).
        The top n rows of A_rem contain Householder reflectors.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update. Together with the top n rows of A_rem,
        this can be used to apply the block Householder transformation to other matrices.
        The number of rows depends on the block size and is architecture-dependent.
    """

@overload
def update_cholesky_sign(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], signs: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def update_cholesky_diag(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], diag: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Cholesky factorization update with diagonal scaling. Returns updated copies.

    L̃L̃ᵀ + ÃDÃᵀ = LLᵀ + ADAᵀ,
    where D = np.diag(diag).

    Parameters
    ----------
    L : (k × n), lower-trapezoidal
        The original Cholesky factor.

    A : (k × m), rectangular
        The update matrix.

    diag : m-vector
        Scale factors corresponding to the columns of A.

    Returns
    -------
    L̃ : (k × n)
        The updated Cholesky factor.

    A_rem : (k × m)
        Contains the k-n bottom rows of the remaining update matrix Ã.
        The top n rows of Ã are zero (not stored explicitly).
        The top n rows of A_rem contain Householder reflectors.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update. Together with the top n rows of A_rem,
        this can be used to apply the block Householder transformation to other matrices.
        The number of rows depends on the block size and is architecture-dependent.
    """

@overload
def update_cholesky_diag(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], diag: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def update_apply_householder(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], B: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Apply a block Householder transformation generated during a Cholesky
    factorization update. Returns updated copies.

    (L̃ Ã) = (L A) Q

    where Q is the block Householder transformation represented by W and B.

    Parameters
    ----------
    L : (l × n), rectangular
        Matrix to apply the transformation to.

    A : (l × m), rectangular
        Matrix to apply the transformation to.

    B : (k × m), rectangular
        The Householder reflector vectors generated during the Cholesky
        factorization update.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update.

    Returns
    -------
    L̃ : (l × n)
        The updated matrix L.

    Ã : (l × m)
        The updated matrix A.
    """

@overload
def update_apply_householder(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], B: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def downdate_apply_householder(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], B: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Apply a block Householder transformation generated during a Cholesky
    factorization downdate. Returns updated copies.

    (L̃ Ã) = (L A) Q

    where Q is the block Householder transformation represented by W and B.

    Parameters
    ----------
    L : (l × n), rectangular
        Matrix to apply the transformation to.

    A : (l × m), rectangular
        Matrix to apply the transformation to.

    B : (k × m), rectangular
        The Householder reflector vectors generated during the Cholesky
        factorization downdate.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization downdate.

    Returns
    -------
    L̃ : (l × n)
        The updated matrix L.

    Ã : (l × m)
        The updated matrix A.
    """

@overload
def downdate_apply_householder(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], B: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def update_apply_householder_sign(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], signs: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)], B: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Apply a block Householder transformation generated during a Cholesky
    factorization update with signed columns. Returns updated copies.

    (L̃ Ã) = (L A) Q

    where Q is the block Householder transformation represented by W, B and signs.

    Parameters
    ----------
    L : (l × n), rectangular
        Matrix to apply the transformation to.

    A : (l × m), rectangular
        Matrix to apply the transformation to.

    signs : m-vector
        Signs that determine whether a column of A was added (+0) or removed (-0).
        Values other than ±0 are not allowed.

    B : (k × m), rectangular
        The Householder reflector vectors generated during the Cholesky
        factorization update.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update.

    Returns
    -------
    L̃ : (l × n)
        The updated matrix L.

    Ã : (l × m)
        The updated matrix A.
    """

@overload
def update_apply_householder_sign(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], signs: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)], B: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...

@overload
def update_apply_householder_diag(L: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float64], dict(shape=(None, None), device='cpu', writable=False)], diag: Annotated[NDArray[numpy.float64], dict(shape=(None,), order='A', device='cpu', writable=False)], B: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float64], dict(shape=(None, None), order='F')]]:
    """
    Apply a block Householder transformation generated during a Cholesky
    factorization update with diagonal scaling. Returns updated copies.

    (L̃ Ã) = (L A) Q

    where Q is the block Householder transformation represented by W, B and diag.

    Parameters
    ----------
    L : (l × n), rectangular
        Matrix to apply the transformation to.

    A : (l × m), rectangular
        Matrix to apply the transformation to.

    diag : m-vector
        Scale factors corresponding to the columns of A used when generating the
        Householder transformation.

    B : (k × m), rectangular
        The Householder reflector vectors generated during the Cholesky
        factorization update.

    W : (r × n)
        The upper triangular Householder representations generated during the
        Cholesky factorization update.

    Returns
    -------
    L̃ : (l × n)
        The updated matrix L.

    Ã : (l × m)
        The updated matrix A.
    """

@overload
def update_apply_householder_diag(L: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], A: Annotated[NDArray[numpy.float32], dict(shape=(None, None), device='cpu', writable=False)], diag: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='A', device='cpu', writable=False)], B: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)], W: Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')], Annotated[NDArray[numpy.float32], dict(shape=(None, None), order='F')]]: ...



import contextlib, importlib, os, sys

_variant = os.getenv("HYHOUND_VARIANT")
if _variant is None:
    _variant = "generic"
    with contextlib.suppress(ModuleNotFoundError):
        from ._dispatch import get_dispatch_name

        _variant = get_dispatch_name()

_target_name = __name__ + "_" + _variant
_target = importlib.import_module(_target_name, package=__package__)
setattr(_target, "variant", _variant)
sys.modules[__name__] = _target



import os

from conan import ConanFile
from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout
from conan.tools.build import can_run
from conan.tools.files import save
from conan.tools.scm import Git


class HyhoundRecipe(ConanFile):
    name = "hyhound"
    version = "1.0.2-alpha.3"

    # Optional metadata
    license = "LGPL-3.0-or-later"
    author = "Pieter P <pieter.p.dev@outlook.com>"
    url = "https://github.com/kul-optec/hyhound"
    description = "Hyperbolic Householder transformations for Cholesky factorization up- and downdates."
    topics = "scientific software"

    # Binary configuration
    package_type = "library"
    settings = "os", "compiler", "build_type", "arch"
    bool_hyhound_options = {
        "with_ocp": False,
        "with_benchmarks": False,
        "with_python": False,
        "with_python_dispatch": False,
    }
    options = {
        "shared": [True, False],
        "fPIC": [True, False],
        "real_type": ["double;float", "float;double", "double", "float"],
        "with_conan_python": [True, False],
    } | {k: [True, False] for k in bool_hyhound_options}
    default_options = {
        "shared": False,
        "fPIC": True,
        "real_type": "double;float",
        "with_conan_python": False,
    } | bool_hyhound_options

    # Sources are located in the same place as this recipe, copy them to the recipe
    exports_sources = (
        "CMakeLists.txt",
        "src/*",
        "cmake/*",
        "test/*",
        "benchmarks/*",
        "LICENSE",
        "README.md",
    )

    def export_sources(self):
        git = Git(self)
        status_cmd = "status . --short --no-branch --untracked-files=no"
        dirty = bool(git.run(status_cmd).strip())
        hash = git.get_commit() + ("-dirty" if dirty else "")
        print("Commit hash:", hash)
        save(self, os.path.join(self.export_sources_folder, "commit.txt"), hash)

    generators = ("CMakeDeps",)

    def requirements(self):
        self.requires(
            "guanaqo/1.0.0-alpha.20", transitive_headers=True, transitive_libs=True
        )
        if self.options.with_ocp:
            self.requires("eigen/5.0.0", transitive_headers=True)
        elif self.options.with_benchmarks:
            self.requires("eigen/5.0.0")
        else:
            self.test_requires("eigen/5.0.0")
        if self.options.with_benchmarks:
            self.requires("benchmark/1.9.4")
        if self.options.with_python:
            self.requires("nanobind/2.9.2")
            if self.options.with_python_dispatch:
                self.requires("cpu_features/0.10.1")
            if self.options.with_conan_python:
                self.requires("tttapa-python-dev/3.13.7")

    def build_requirements(self):
        self.test_requires("gtest/1.17.0")
        self.tool_requires("cmake/[>=3.24 <4.2]")

    def config_options(self):
        if self.settings.get_safe("os") == "Windows":
            self.options.rm_safe("fPIC")

    def configure(self):
        with_tests = not self.conf.get("tools.build:skip_test", default=False)
        if self.options.with_benchmarks or self.options.with_ocp or with_tests:
            self.options["guanaqo/*"].with_blas = True

    def layout(self):
        cmake_layout(self)
        self.cpp.build.builddirs.append("")

    def generate(self):
        tc = CMakeToolchain(self)
        for k in self.bool_hyhound_options:
            value = self.options.get_safe(k)
            if value is not None and value.value is not None:
                tc.variables["HYHOUND_" + k.upper()] = bool(value)
        guanaqo = self.dependencies["guanaqo"]
        index_type = guanaqo.options.get_safe("blas_index_type", default="int")
        real_type = str(self.options.real_type)
        print("index_type:", index_type)
        print("real_type: ", real_type)
        tc.variables["HYHOUND_DENSE_INDEX_TYPE"] = index_type
        tc.variables["HYHOUND_DENSE_REAL_TYPE"] = real_type
        if can_run(self):
            tc.variables["HYHOUND_FORCE_TEST_DISCOVERY"] = True
        tc.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()
        cmake.test()

    def package(self):
        cmake = CMake(self)
        cmake.install()

    def package_info(self):
        self.cpp_info.set_property("cmake_find_mode", "none")
        self.cpp_info.builddirs.append(os.path.join("lib", "cmake", "hyhound"))
